# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import hashlib
import logging
import os
import random
import string
import time
import yaml

from Crypto.Hash import MD5, SHA256
from Crypto.PublicKey import RSA

from libcloud.compute.providers import get_driver as get_compute_driver
from libcloud.compute.types import Provider as ComputeProvider
from libcloud.loadbalancer.providers import get_driver as get_balancer_driver
from libcloud.loadbalancer.types import Provider as BalancerProvider

from exception import PlumberyException
from facility import PlumberyFacility
from facility import PlumberyFittings
from polisher import PlumberyPolisher
from plumbery import __version__

__all__ = ['PlumberyEngine']


class PlumberyEngine(object):
    """
    Infrastructure as code at Dimension Data with Apache Libcloud

    :param fileName: the location of the plan for the fittings
    :type fileName: ``str``

    Plumbery is a convenient tool for infrastructure managers at cloud times.
    It allows for easy and repeatable deployments of various
    fittings, including compute nodes and related storage. It allows also for
    quick dismandling of the fittings when required.
    The idea here is not to compete with respected solutions such as chef or
    puppet. At its name implies, plumbery is targeting pipes and fittings, the
    very basic utility stuff that sophisticated tools can leverage.

    Example::

        from plumbery.engine import PlumberyEngine
        engine = PlumberyEngine('fittings.yaml')
        engine.build_all_blueprints()

    In this example the overall plan, in textual format, is given to the engine
    in preparation of subsequent processing. The software is not trying to
    guess a name by default, so if you do not provide a name, no configuration
    file is loaded. You can load the plan at any stage, or restart the engine
    with an updated plan, by invoking the member function ``parse_layout()``

    Note:
        While plumbery is not making assumptions for your configuration files,
        if your infrastructure is simple enough to fit in one single file then
        you are highly encouraged to name it ``fittings.yaml``

    Beyond the plan for your fittings, plumbery is also requiring some specific
    credentials to connect to cloud providers. To preserve the confidentiality
    of such information, it is read from the environment, and not from any
    configuration file. Therefore the need for local setup before running
    plumbery. This is part of the installation process.

    Last but not least, plumbery sets the root password of any new server that
    it creates. For obvious security reasons this is not taken from the
    fittings plan but from the environment, or it can be set in code.

    Under Linux, you may want to edit ``~/.bash_profile`` like this::

        # credentials to access cloud resources from Dimension Data
        export MCP_USERNAME='foo.bar'
        export MCP_PASSWORD='WhatsUpDoc'

        # password to access nodes remotely
        export SHARED_SECRET='*you really want to put a tricky password here*'

    These simple precautions are aiming to protect your infrastructure as much
    as possible. The general goal is to minimize risks induced by exposure to
    your fittings plan. You may lead transformation towards so-called
    infrastructure as code, and for this you will add version control to your
    fittings plan. As a result, plans will be stored in git or equivalent, and
    shared across some people.

    Attributes:

        safeMode (boolean):
            If True, which is the default, then no actual change
            will be made against the infrastructure. This global attribute
            is coming from the fittings plan.

    """

    safeMode = True

    def __init__(self, plan=None):
        """
        Ignites the plumbing engine

        :param plan: the file that contains fittings plan
        :type plan: ``str`` or ``file``

        """

        self.fittingsFile = None

        self.facilities = []

        self.polishers = []

        self._buildPolisher = None

        self._sharedSecret = None

        self.secrets = {}

        self._userName = None

        self._userPassword = None

        if plan is not None:
            self.from_file(plan)

    def from_file(self, plan=None):
        """
        Reads the fittings plan from a file

        :param plan: the file that contains fittings plan
        :type plan: ``str`` or ``file``

        The fittings plan is expected to follow YAML specifications, and it
        must have multiple documents in it. The first document provides
        general configuration parameters for the engine. Subsequent documents
        describe the various locations for the fittings.

        An example of a minimum fittings plan::

            ---
            safeMode: False
            ---
            # Frankfurt in Europe
            locationId: EU6
            regionId: dd-eu

            blueprints:

              - myBluePrint:
                  domain:
                    name: myDC
                  ethernet:
                    name: myVLAN
                    subnet: 10.1.10.0
                  nodes:
                    - myServer

        In this example, the plan is to deploy a single node in the data centre
        at Frankfurt, in Europe. The node `myServer` will be placed in a
        network named `myVLAN`, and the network will be part of a network
        domain acting as a virtual data centre, `myDC`. The blueprint has a
        name, `myBluePrint`, so that it can be handled independently from
        other blueprints.

        """

        if plan is None:
            plan = os.getenv('PLUMBERY')

        if isinstance(plan, str):
            self.fittingsFile = plan
            plan = open(plan, 'r')

        documents = yaml.load_all(plan)

        # first document contains engine settings
        self.set(documents.next())

        # then one document per facility
        for document in documents:
            self.add_facility(document)

        self.load_secrets()

        if self.safeMode:
            logging.info(
                "Running in safe mode"
                " - no actual change will be made to the fittings")

    def from_text(self, plan):
        """
        Reads the fittings plan

        :param plan: the fittings plan
        :type plan: ``str``

        The fittings plan is expected to follow YAML specifications, and it
        must have multiple documents in it. The first document provides
        general configuration parameters for the engine. Subsequent documents
        describe the various locations for the fittings.

        Example of use::

            myPlan = \"\"\"
            ---
            safeMode: True
            ---
            # Frankfurt in Europe
            locationId: EU6
            regionId: dd-eu

            blueprints:

              - myBlueprint:
                  domain:
                    name: myDC
                  ethernet:
                    name: myVLAN
                    subnet: 10.1.10.0
                  nodes:
                    - myServer
            \"\"\"

            engine = PlumberyEngine()
            engine.from_text(myPlan)

        In this example, the plan is to deploy a single node in the data centre
        at Frankfurt, in Europe. The node `myServer` will be placed in a
        network named `myVLAN`, and the network will be part of a network
        domain acting as a virtual data centre, `myDC`. The blueprint has a
        name, `myBluePrint`, so that it can be handled independently from
        other blueprints.

        """

        self.from_file(io.TextIOWrapper(io.BytesIO(plan)))

    def set(self, settings):
        """
        Changes the settings of the engine

        :param settings: the new settings
        :type settings: ``dict``

        """

        if not isinstance(settings, dict):
            raise TypeError('settings should be a dictionary')

        if 'safeMode' not in settings:
            raise LookupError('safeMode is not defined')

        self.safeMode = settings['safeMode']
        if self.safeMode not in [True, False]:
            raise ValueError('safeMode should be either True or False')

        if 'polishers' in settings:
            for item in settings['polishers']:
                key = item.keys()[0]
                value = item[key]
                self.polishers.append(
                    PlumberyPolisher.from_shelf(key, value))

        if 'buildPolisher' in settings:
            self._buildPolisher = settings['buildPolisher']
        else:
            self._buildPolisher = 'spit'

    def set_shared_secret(self, secret):
        """
        Changes the shared secret to be used with new nodes

        :param secret: the user name to be used with the driver
        :type secret: ``str``

        This function can be used to supplement the normal provision of
        the shared secret via the environment variable ``SHARED_SECRET``.

        """

        self._sharedSecret = secret

    def get_shared_secret(self):
        """
        Retrieves the secret that is communicated to new nodes during setup

        :return: the shared secret to be given to the driver
        :rtype: ``str``

        :raises: :class:`plumbery.PlumberyException`
            - if no shared secret can be found

        The shared secret is not put in the fittings plan, but is normally
        taken from the environment variable ``SHARED_SECRET``.

        Under Linux, you may want to edit ``~/.bash_profile`` like this::

            # password to access nodes remotely
            export SHARED_SECRET='*you really want to use a tricky password*'

        Alternatively, you can use the member function ``set_shared_secret()``
        to set this important attribute via code.

        """

        if self._sharedSecret is None:
            self._sharedSecret = os.getenv('SHARED_SECRET')
            if self._sharedSecret is None or len(self._sharedSecret) < 3:
                raise PlumberyException(
                    "Error: missing node password "
                    "in environment SHARED_SECRET")

        return self._sharedSecret

    def get_rsa_secret(self, id='pair.rsa_private'):
        """
        Returns a part of a RSA pair of keys

        :param id: name of the key
        :type id: ``str``

        """

        if id in self.secrets:
            return self.secrets[id]

        name = id.split('.')[0]

        key = RSA.generate(2048)
        self.secrets[name+'.rsa_private'] = key.exportKey('PEM')
        logging.debug("- generating {} -> {}".format(
            name+'.rsa_private', self.secrets[name+'.rsa_private']))

        pubkey = key.publickey()
        self.secrets[name+'.rsa_public'] = pubkey.exportKey('PEM')
        logging.debug("- generating {} -> {}".format(
            name+'.rsa_public', self.secrets[name+'.rsa_public']))

        self.secrets[name+'.ssh.rsa_public'] = pubkey.exportKey('OpenSSH')
        logging.debug("- generating {} -> {}".format(
            name+'.ssh.rsa_public', self.secrets[name+'.ssh.rsa_public']))

        self.save_secrets()
        return self.secrets[id]

    def get_secret(self, id='random'):
        """
        Returns a secret

        :param id: name of this secret
        :type id: ``str``

        :return: a transient random secret
        :rtype: ``str``

        Random secrets can be used in scripts and in configuration files
        sent to nodes, for example to configure a database server.

        For this you would put ``{{ random.secret }}`` in your files and let
        plumbery provide a value for you.

        The `id` parameter designates one secret among several.

        The `format` parameter specifies the kind of string that is expected:
        * 'user' - to be read by a human being
        * 'sha1' - rather long string too
        * 'md5' - rather long string
        * 'sha256' - longest and most secure

        Some examples:

            {{ master.md5.secret }}
            {{ slave.secret }}
            {{ server357.sha1.secret }}

        In a nutshell, this function is giving you a lot of flexibility
        in the generation of secrets.

        """

        if id in self.secrets:
            return self.secrets[id]

        secret = ''.join(random.choice(
            string.ascii_letters+string.digits+'-_!=')
                for i in range(50))

        if id.endswith('.sha256.secret'):
            secret = SHA256.new(secret).hexdigest()

        elif id.endswith('.md5.secret'):
            secret = MD5.new(secret).hexdigest()

        elif id.endswith('.sha1.secret'):
            secret = hashlib.sha1(secret).hexdigest()

        else:
            secret = secret[0:9]

        logging.debug("- generating {} -> {}".format(id, secret))
        self.secrets[id] = secret
        self.save_secrets()

        return secret

    def save_secrets(self):
        """
        Saves secrets attached to this fittings plan
        """

        if self.fittingsFile is not None:

            if self.fittingsFile.endswith('.yaml'):
                secretsFile = self.fittingsFile[0:-len('.yaml')]+'.secrets'
            else:
                secretsFile = self.fittingsFile+'.secrets'

            try:
                handle = open(secretsFile, 'w')
                for id in self.secrets:
                    handle.write("{}: '{}'\n".format(
                        id, self.secrets[id].replace('\n', '\\n')))
                handle.close()

            except IOError:
                logging.warning("Unable to write secrets to '{}'".format(
                    secretsFile))

    def load_secrets(self):
        """
        Loads secrets attached to this fittings plan
        """

        if self.fittingsFile is not None:

            if self.fittingsFile.endswith('.yaml'):
                secretsFile = self.fittingsFile[0:-len('.yaml')]+'.secrets'
            else:
                secretsFile = self.fittingsFile+'.secrets'

            try:
                handle = open(secretsFile, 'r')
                self.secrets = yaml.load(handle)
                handle.close()

                logging.debug("Loading {} secrets from '{}'".format(
                    len(self.secrets), secretsFile))

            except IOError:
                logging.debug("Unable to load secrets from '{}'".format(
                    secretsFile))

    def list_secrets(self):
        """
        Lists secrets attached to this fittings plan
        """

        logging.info("Listing secrets")

        if len(self.secrets.keys()) < 1:
            logging.info("- no secret found")

        for key in self.secrets.keys():
            logging.info("- {}: {}".format(key, self.secrets[key]))

    def set_user_name(self, name):
        """
        Changes the name used to authenticate to the API

        :param name: the user name to be used with the driver
        :type name: ``str``

        This function can be used to supplement the normal provision of
        a user name via the environment variable ``MCP_USERNAME``.

        """

        self._userName = name

    def get_user_name(self):
        """
        Retrieves user name to authenticate to the API

        :return: the user name to be used with the driver
        :rtype: ``str``

        :raises: :class:`plumbery.PlumberyException`
            - if no user name can be found

        The user name is not put in the fittings plan, but is normally taken
        from the environment variable ``MCP_USERNAME``.

        Under Linux, you may want to edit ``~/.bash_profile`` like this::

            # credentials to access cloud resources from Dimension Data
            export MCP_USERNAME='foo.bar'
            export MCP_PASSWORD='WhatsUpDoc'

        """

        if self._userName is None:
            self._userName = os.getenv('MCP_USERNAME')
            if self._userName is None or len(self._userName) < 3:
                raise PlumberyException(
                    "Error: missing credentials in environment MCP_USERNAME")

        return self._userName

    def set_user_password(self, password):
        """
        Changes the password used to authenticate to the API

        :param password: the user password to be used with the driver
        :type password: ``str``

        This function can be used to supplement the normal provision of
        a user password via the environment variable ``MCP_PASSWORD``.

        """

        self._userPassword = password

    def get_user_password(self):
        """
        Retrieves user password to authenticate to the API

        :return: the user password to be used with the driver
        :rtype: ``str``

        :raises: :class:`plumbery.PlumberyException`
            - if no user password can be found

        The user password is not put in the fittings plan, but is normally
        taken from the environment variable ``MCP_PASSWORD``.

        Under Linux, you may want to edit ``~/.bash_profile`` like this::

            # credentials to access cloud resources from Dimension Data
            export MCP_USERNAME='foo.bar'
            export MCP_PASSWORD='WhatsUpDoc'

        """

        if self._userPassword is None:
            self._userPassword = os.getenv('MCP_PASSWORD')
            if self._userPassword is None or len(self._userPassword) < 3:
                raise PlumberyException(
                    "Error: missing credentials in environment MCP_PASSWORD")

        return self._userPassword

    def add_facility(self, facility):
        """
        Extends the scope of this plumbing engine

        :param facility: description of an additional facility
        :type facility: ``dict`` or :class:`plumbery.PlumberyFacility`

        """

        if isinstance(facility, dict):
            facility = PlumberyFacility(self, PlumberyFittings(**facility))

        logging.debug("Adding facility")
        logging.debug(facility.fittings)

        self.facilities.append(facility)

    def list_facility(self, location):
        """
        Retrieves facilities by their location

        :param location: the target location, e.g., 'EU6'
        :type location: ``str`` or ``list`` of ``str``

        :return: the list of matching facilities
        :rtype: ``list`` of :class:`plumbery.PlumberyFacility` or ``[]``

        Examples::

            >>>engine.list_facility(location='EU6')
            ...

            >>>engine.list_facility(location='EU6 NA9')
            ...

            >>>engine.list_facility(location=['EU6', 'NA9'])
            ...

        """

        if isinstance(location, str):
            location = location.split(' ')

        matches = []

        for item in location:
            if isinstance(item, PlumberyFacility):
                matches.append(item)

        for facility in self.facilities:
            if facility.fittings.locationId in location:
                matches.append(facility)

        return matches

    def do(self, action, blueprints=None, facilities=None):
        """
        Applies an action to multiple blueprints at multiple locations

        :param action: 'deploy', 'dispose', etc., or a polisher name
        :type action: ``str``

        :param blueprints: a list of blueprints
        :type blueprints: ``list`` of ``str``, or None

        :param facilities: a list of facilities
        :type factilities: ``list`` of ``str``, or None

        """

        if action == 'secrets':
            self.list_secrets()

        elif action == 'deploy':
            if blueprints is None:
                self.build_all_blueprints(facilities)
                self.start_all_blueprints(facilities)
                self.polish_all_blueprints(filter='rub',
                                           facilities=facilities)
                self.polish_all_blueprints(filter='information',
                                           facilities=facilities)
            else:
                self.build_blueprint(blueprints, facilities)
                self.start_blueprint(blueprints, facilities)
                self.polish_blueprint(blueprints,
                                      filter='rub',
                                      facilities=facilities)
                self.polish_blueprint(blueprints,
                                      filter='information',
                                      facilities=facilities)

        elif action == 'dispose':
            if blueprints is None:
                self.stop_all_blueprints(facilities)
                self.destroy_all_blueprints(facilities)
            else:
                self.stop_blueprint(blueprints, facilities)
                self.destroy_blueprint(blueprints, facilities)

        elif action == 'build':
            if blueprints is None:
                self.build_all_blueprints(facilities)
            else:
                self.build_blueprint(blueprints, facilities)

        elif action == 'start':
            if blueprints is None:
                self.start_all_blueprints(facilities)
            else:
                self.start_blueprint(blueprints, facilities)

        elif action == 'polish':
            if blueprints is None:
                self.polish_all_blueprints(filter=None,
                                           facilities=facilities)
            else:
                self.polish_blueprint(blueprints,
                                      filter=None,
                                      facilities=facilities)

        elif action == 'stop':
            if blueprints is None:
                self.stop_all_blueprints(facilities)
            else:
                self.stop_blueprint(blueprints, facilities)

        elif action == 'wipe':
            if blueprints is None:
                self.wipe_all_blueprints(facilities)
            else:
                self.wipe_blueprint(blueprints, facilities)

        elif action == 'destroy':
            if blueprints is None:
                self.destroy_all_blueprints(facilities)
            else:
                self.destroy_blueprint(blueprints, facilities)

        else:
            if blueprints is None:
                self.polish_all_blueprints(filter=action,
                                           facilities=facilities)
            else:
                self.polish_blueprint(blueprints,
                                      filter=action,
                                      facilities=facilities)

    def build_all_blueprints(self, facilities=None):
        """
        Builds all blueprints described in fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to build all blueprints there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Examples::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_all_blueprints()

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_all_blueprints('EU6 NA9')

        """

        logging.info("Building all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.build_all_blueprints()

        self.polish_all_blueprints(filter=self._buildPolisher,
                                   facilities=facilities)

    def build_blueprint(self, names, facilities=None):
        """
        Builds named blueprint from fittings plan

        :param names: the name(s) of the blueprint(s) to deploy
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to build one single blueprint there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_blueprints('sql')

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        logging.info("Building blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.build_blueprint(names)

        self.polish_blueprint(names=names,
                              filter=self._buildPolisher,
                              facilities=facilities)

    def start_all_blueprints(self, facilities=None):
        """
        Starts all nodes described in the fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to start all nodes there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already up and running.

        """

        logging.info("Starting nodes from all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.start_all_blueprints()

    def start_blueprint(self, names, facilities=None):
        """
        Starts nodes of one blueprint of the fittings plan

        :param names: the name(s) of the blueprint(s) to start
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to start nodes from some blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already up and running.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        logging.info("Starting nodes from blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.start_blueprint(names)

    def polish_all_blueprints(self, filter=None, facilities=None):
        """
        Walks all resources and polishes them

        :param filter: the name of a single polisher to apply. If this
            parameter is missing, all polishers declared in the fittings plan
            will be applied
        :type filter: ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to apply polishers there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').polish_all_blueprints()

        """

        polishers = PlumberyPolisher.filter(self.polishers, filter)

        if len(polishers) < 1:
            return False

        logging.info("Polishing all blueprints")

        for polisher in polishers:
            polisher.go(self)

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            for polisher in polishers:
                polisher.move_to(facility)
            facility.polish_all_blueprints(polishers)

        for polisher in polishers:
            polisher.reap()

    def polish_blueprint(self, names, filter=None, facilities=None):
        """
        Walks resources from the target blueprint and polishes them

        :param names: the name(s) of the blueprint(s) to polish
        :type names: ``str`` or ``list`` of ``str``

        :param filter: the name of a single polisher to apply. If this
            parameter is missing, all polishers declared in the fittings plan
            will be applied
        :type filter: ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to apply one polisher to one blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').polish_blueprint('sql')

        """

        polishers = PlumberyPolisher.filter(self.polishers, filter)

        if len(polishers) < 1:
            logging.debug('No polisher has been found')
            return

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        logging.info("Polishing blueprint '{}'".format(label))

        for polisher in polishers:
            polisher.go(self)

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            for polisher in polishers:
                polisher.move_to(facility)
            facility.polish_blueprint(names, polishers)

        for polisher in polishers:
            polisher.reap()

    def stop_all_blueprints(self, facilities=None):
        """
        Stops all nodes of the fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to stop all nodes there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already stopped.

        """

        logging.info("Stopping nodes from all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.stop_all_blueprints()

    def stop_blueprint(self, names, facilities=None):
        """
        Stops nodes of one blueprint of the fittings plan

        :param names: the name(s) of the blueprint(s) to stop
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to stop nodes from some blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        This function has no effect on nodes that are already stopped.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        logging.info("Stopping nodes from blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.stop_blueprint(names)

    def wipe_all_blueprints(self, facilities=None):
        """
        Destroys all nodes from fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy all nodes there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        logging.info("Wiping all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.wipe_all_blueprints()

    def wipe_blueprint(self, names, facilities=None):
        """
        Destroys nodes for one or several blueprint(s) of the fittings plan

        :param names: the name(s) of the blueprint(s) to destroy
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy nodes from one blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        logging.info("Wiping blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.wipe_blueprint(names)

    def destroy_all_blueprints(self, facilities=None):
        """
        Destroys all blueprints from fittings plan

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy all blueprints there.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        logging.info("Destroying all blueprints")

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.destroy_all_blueprints()

    def destroy_blueprint(self, names, facilities=None):
        """
        Destroys one or several blueprint(s) from fittings plan

        :param names: the name(s) of the blueprint(s) to destroy
        :type names: ``str`` or ``list`` of ``str``

        :param facilities: explicit list of target facilities
        :type facilities: ``str`` or ``list`` of ``str``

        This function checks facilities to destroy one single blueprint.
        The default behaviour is to consider all facilities mentioned in the
        fittings plan. If a list of facilities is provided, than the action is
        limited to this subset only.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        if isinstance(names, list):
            label = ' '.join(names)
        else:
            label = names

        logging.info("Destroying blueprint '{}'".format(label))

        if facilities is not None:
            facilities = self.list_facility(facilities)
        else:
            facilities = self.facilities

        for facility in facilities:
            facility.focus()
            facility.destroy_blueprint(names)

    def get_compute_driver(self, region):
        """
        Loads a compute driver from Apache Libcloud

        """

        driver = get_compute_driver(ComputeProvider.DIMENSIONDATA)

        return driver(
            key=self.get_user_name(),
            secret=self.get_user_password(),
            region=region)

    def get_balancer_driver(self, region):
        """
        Loads a load balancer driver from Apache Libcloud

        """

        driver = get_balancer_driver(BalancerProvider.DIMENSIONDATA)

        return driver(
            key=self.get_user_name(),
            secret=self.get_user_password(),
            region=region)

    def lookup(self, token):

        if token == 'plumbery.version':
            return __version__

        if token == 'shared.secret':
            return self.get_shared_secret()

        if token.endswith('.secret'):
            return self.get_secret(token)

        if token.endswith('.rsa_public'):
            return self.get_rsa_secret(token)
        if token.endswith('.rsa_private'):
            return self.get_rsa_secret(token)

        return None




# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from uuid import uuid4

try:
    from lxml import etree as ET
except ImportError:
    from xml.etree import ElementTree as ET

from libcloud.common.dimensiondata import DimensionDataFirewallRule
from libcloud.common.dimensiondata import DimensionDataFirewallAddress
from libcloud.common.dimensiondata import DimensionDataNatRule
from libcloud.common.dimensiondata import TYPES_URN
from libcloud.loadbalancer.base import Algorithm
from libcloud.loadbalancer.base import Member
from libcloud.loadbalancer.types import State

from libcloud.utils.xml import fixxpath, findtext, findall

from exception import PlumberyException

__all__ = ['PlumberyInfrastructure']


class PlumberyInfrastructure(object):
    """
    Infrastructure as code, for network and security

    :param facility: the underlying physical facility
    :type facility: :class:`plumbery.PlumberyFacility`

    This is an abstraction of a virtual data center. It is a secured
    container for multiple nodes.

    Example::

        from plumbery.infrastructure import PlumberyInfrastructure
        infrastructure = PlumberyInfrastructure(facility)
        infrastructure.build(blueprint)

    In this example an infrastructure is initialised at the given facility, and
    then it is asked to create pipes and plumbing described in the
    provided blueprint. This is covering solely the network and the security,
    not the nodes themselves.

    Attributes:
        facility (PlumberyFacility):
            a handle to the physical facility where network domains
            are implemented

    """

    # the physical data center
    facility = None

    def __init__(self, facility=None):
        """A virtual data centre attached to a physical data centre"""

        # handle to parent parameters and functions
        self.facility = facility
        self.region = facility.region
        self.plumbery = facility.plumbery
        self.network = None
        self.domain = None

        self._cache_remote_vlan = []
        self._cache_offshore_vlan = []
        self._cache_firewall_rules = []
        self._cache_listeners = None
        self._cache_pools = None

        self._network_domains_already_built = []
        self._vlans_already_built = []

    def get_region_id(self):
        return self.facility.fittings.regionId

    def _attach_node(self, node, networks):
        """
        Glues a node to multiple networks

        :param node: the target node
        :type node: :class:`libcloud.compute.base.Node`

        :param networks: a list of networks to connect, and ``internet``
        :type networks: list of ``str``

        This function adds network interfaces to a node, or adds address
        translation to the public Internet.

        Example in the fittings plan::

          - web:
              domain:
                ipv4: 6
              ethernet:
                name: gigafox.data
              nodes:
                - web[10..12]:
                    glue:
                      - gigafox.control
                      - internet 80 443

        In this example, another network interface is added to each node for
        connection to the Ethernet network ``gigafox.control``.

        Also, public IPv4 addresses are mapped on private addresses, so that
        each node web10, web11 and web12 is reachable from the internet.
        Public IPv4 addresses are taken from pool declared at the domain level,
        with the attribute ``ipv4``. In the example above, 6 addresses are
        assigned to the network domain, of which 3 are given to web nodes.

        If one or multiple numbers are mentioned after the network name, they
        are used to configure the firewall appropriately.

        """

        hasChanged = False

        if node is None:
            return hasChanged

        for line in networks:

            tokens = line.split(' ')
            label = tokens.pop(0)

            if self.plumbery.safeMode:
                logging.info("Glueing node '{}' to network '{}'"
                             .format(node.name, label))
                logging.info("- not in safe mode")
                continue

            if label == 'internet':
                self._attach_node_to_internet(node, tokens)
                continue

            logging.info("Glueing node '{}' to network '{}'"
                         .format(node.name, label))
            target = self.get_ethernet(label.split('::'))
            if target is None:
                logging.info("- network '{}' is unknown".format(label))
                continue

            while True:
                try:
                    self.region.ex_attach_node_to_vlan(node, target)
                    logging.info("- in progress")
                    hasChanged = True

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        logging.info("- not now - locked")

                    elif 'INVALID_INPUT_DATA' in str(feedback):
                        logging.info("- already there")

                    else:
                        logging.info("- unable to glue node")
                        logging.error(str(feedback))

                break

        return hasChanged

    def _attach_node_to_internet(self, node, ports=[]):
        """
        Adds address translation for one node

        :param node: node that has to be reachable from the internet
        :type node: :class:`libcloud.common.Node`

        :param ports: the ports that have to be opened
        :type ports: a list of ``str``

        """

        domain = self.get_network_domain(self.blueprint['domain']['name'])

        internal_ip = node.private_ips[0]

        external_ip = None
        for rule in self.region.ex_list_nat_rules(domain):
            if rule.internal_ip == internal_ip:
                external_ip = rule.external_ip
                logging.info("Making node '{}' reachable from the internet"
                             .format(node.name))
                logging.info("- node is reachable at '{}'".format(external_ip))

        if external_ip is None:
            external_ip = self._get_ipv4()

            if external_ip is None:
                logging.info("Making node '{}' reachable from the internet"
                             .format(node.name))
                logging.info("- no more ipv4 address available -- assign more")
                return

            logging.info("Making node '{}' reachable from the internet"
                         .format(node.name))
            while True:
                try:
                    self.region.ex_create_nat_rule(
                        domain,
                        internal_ip,
                        external_ip)
                    logging.info("- node is reachable at '{}'".format(
                        external_ip))

                except Exception as feedback:
                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        logging.info("- not now - locked")
                        return

                    else:
                        logging.info("- unable to add address translation")
                        logging.error(str(feedback))

                break

        candidates = self._list_candidate_firewall_rules(node, ports)

        for rule in self._list_firewall_rules():

            if rule.name in candidates.keys():
                logging.info("Creating firewall rule '{}'"
                             .format(rule.name))
                logging.info("- already there")
                candidates = {k: candidates[k]
                              for k in candidates if k != rule.name}

        for name, rule in candidates.items():

            logging.info("Creating firewall rule '{}'"
                         .format(name))

            if self.plumbery.safeMode:
                logging.info("- not in safe mode")

            else:

                try:

                    self._ex_create_firewall_rule(
                        network_domain=domain,
                        rule=rule,
                        position='LAST')

                    logging.info("- in progress")

                except Exception as feedback:

                    if 'NAME_NOT_UNIQUE' in str(feedback):
                        logging.info("- already there")

                    else:
                        logging.info("- unable to create firewall rule")
                        logging.error(str(feedback))

    def build(self, blueprint):
        """
        Creates the infrastructure for one blueprint

        :param blueprint: the various attributes of the target fittings
        :type blueprint: ``dict``

        :return: ``True`` if the network has been created or is already there,
            ``False`` otherwise
        :rtype: ``bool``

        :raises: :class:`plumbery.PlumberyException`
            - if some unrecoverable error occurs

        This function is looking at all fittings in the blueprint except the
        nodes. This is including:

        * a network domain
        * one Ethernet network
        * eventually, several public IPv4 addresses
        * address translation rules to private IPv4 addresses
        * firewall rules

        In safe mode, the function will stop on any missing component since
        it is not in a position to add fittings, and return ``False``.
        If all components already exist then the funciton will return ``True``.

        """

        self.blueprint = blueprint

        logging.debug("Building infrastructure for blueprint '{}'".format(
            blueprint['target']))

        if 'domain' not in blueprint or type(blueprint['domain']) is not dict:
            raise PlumberyException(
                "Error: no network domain has been defined "
                "for the blueprint '{}'!".format(blueprint['target']))

        domainName = blueprint['domain']['name']

        if 'ethernet' not in blueprint                                      \
           or type(blueprint['ethernet']) is not dict:
            raise PlumberyException(
                "Error: no ethernet network has been defined "
                "for the blueprint '{}'!".format(blueprint['target']))

        if 'subnet' not in blueprint['ethernet']:
            raise PlumberyException(
                "Error: no IPv4 subnet "
                "(e.g., '10.0.34.0') as been defined for the blueprint '{}'!"
                .format(blueprint['target']))

        networkName = blueprint['ethernet']['name']

        self.domain = self.get_network_domain(domainName)
        if self.domain is not None:
            logging.info("Creating network domain '{}'".format(domainName))
            logging.info("- already there")

        elif self.plumbery.safeMode:
            logging.info("Creating network domain '{}'".format(domainName))
            logging.info("- not in safe mode")
            logging.info("Creating Ethernet network '{}'"
                         .format(networkName))
            logging.info("- not in safe mode")
            return False

        else:
            logging.info("Creating network domain '{}'".format(domainName))

            # the description attribute is a smart way to tag resources
            description = '#plumbery'
            if 'description' in blueprint['domain']:
                description = blueprint['domain']['description']+' #plumbery'

            # level of service
            service = 'ESSENTIALS'
            if 'service' in blueprint['domain']:
                service = blueprint['domain']['service'].upper()

            while True:
                try:
                    self.domain = self.region.ex_create_network_domain(
                        location=self.facility.location,
                        name=domainName,
                        service_plan=service,
                        description=description)
                    logging.info("- in progress")

                    # prevent locks in xops
                    self.region.ex_wait_for_state(
                        'NORMAL', self.region.ex_get_network_domain,
                        poll_interval=5, timeout=1200,
                        network_domain_id=self.domain.id)

                    self.facility._cache_network_domains.append(self.domain)

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'OPERATION_NOT_SUPPORTED' in str(feedback):
                        logging.info("- operation not supported")
                        return False

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        logging.info("- not now - locked")
                        return False

                    else:
                        logging.info("- unable to create network domain")
                        logging.error(str(feedback))
                        return False

                break

        self.network = self.get_ethernet(networkName)
        if self.network is not None:
            logging.info("Creating Ethernet network '{}'"
                         .format(networkName))
            logging.info("- already there")

        elif self.plumbery.safeMode:
            logging.info("Creating Ethernet network '{}'"
                         .format(networkName))
            logging.info("- not in safe mode")
            return False

        else:
            logging.info("Creating Ethernet network '{}'"
                         .format(networkName))

            # the description attribute is a smart way to tag resources
            description = '#plumbery'
            if 'description' in blueprint['ethernet']:
                description = blueprint['ethernet']['description']+' #plumbery'

            while True:
                try:
                    self.network = self.region.ex_create_vlan(
                        network_domain=self.domain,
                        name=networkName,
                        private_ipv4_base_address=blueprint['ethernet']['subnet'],
                        description=description)
                    logging.info("- in progress")

                    # prevent locks in xops
                    self.region.ex_wait_for_state(
                        'NORMAL',
                        self.region.ex_get_vlan,
                        poll_interval=5, timeout=1200,
                        vlan_id=self.network.id)

                    self.facility._cache_vlans.append(self.network)

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'NAME_NOT_UNIQUE' in str(feedback):
                        logging.info("- not possible "
                                     "- network already exists elsewhere")

                    elif 'IP_ADDRESS_NOT_UNIQUE' in str(feedback):
                        logging.info("- not possible "
                                     "- subnet is used elsewhere")

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        logging.info("- not now - locked")
                        return False

                    else:
                        logging.info("- unable to create Ethernet network")
                        logging.error(str(feedback))
                        return False

                break

        return True

    def _build_firewall_rules(self):
        """
        Changes firewall settings to accept incoming traffic

        This function adds firewall rules to allow traffic towards given
        network. It looks at the ``accept`` settings in the blueprint to
        identify all source networks.

        Example in the fittings plan::

          - web:
              domain: *vdc1
              ethernet:
                name: gigafox.production
                accept:
                  - gigafox.control
                  - dd-eu::EU6::other.network.there

        In this example, the firewall is configured so that any ip traffic
        from the Ethernet network ``gigafox.control`` can reach the Ethernet
        network ``gigafox.production``. One rule is created for
        IPv4 and another rule is created for IPv6.

        The second network that is configured is from another data centre
        in another region. This is leveraging the private network that
        interconnect all MCPs. For networks outside the current domain, only
        one rule is added to allow IPv6 traffic. This is because IPv4 routing
        is not allowed across multiple network domains.

        """

        if 'accept' not in self.blueprint['ethernet']:
            return True

        destination = self.get_ethernet(self.blueprint['ethernet']['name'])
        if destination is None:
            return True

        destinationIPv4 = DimensionDataFirewallAddress(
            any_ip=False,
            ip_address=destination.private_ipv4_range_address,
            ip_prefix_size=destination.private_ipv4_range_size,
            port_begin=None,
            port_end=None)

        destinationIPv6 = DimensionDataFirewallAddress(
            any_ip=False,
            ip_address=destination.ipv6_range_address,
            ip_prefix_size=destination.ipv6_range_size,
            port_begin=None,
            port_end=None)

        for item in self.blueprint['ethernet']['accept']:

            if isinstance(item, dict):
                label = item.keys()[0]
            else:
                label = str(item)

            source = self.get_ethernet(label.split('::'))
            if source is None:
                logging.info("Source network '{}' is unknown".format(label))
                continue

            ruleIPv4Name = self.name_firewall_rule(
                source.name, destination.name, 'IP')

            shouldCreateRuleIPv4 = True
            if source.location.name != destination.location.name:
                shouldCreateRuleIPv4 = False
            elif source.network_domain.name != destination.network_domain.name:
                shouldCreateRuleIPv4 = False

            ruleIPv6Name = self.name_firewall_rule(
                source.name, destination.name, 'IPv6')

            shouldCreateRuleIPv6 = True

            for rule in self._list_firewall_rules():

                if (shouldCreateRuleIPv4
                        and rule.name.lower() == ruleIPv4Name.lower()):

                    logging.info("Creating firewall rule '{}'"
                                 .format(rule.name))
                    logging.info("- already there")
                    shouldCreateRuleIPv4 = False
                    continue

                if (shouldCreateRuleIPv6
                        and rule.name.lower() == ruleIPv6Name.lower()):

                    logging.info("Creating firewall rule '{}'"
                                 .format(rule.name))
                    logging.info("- already there")
                    shouldCreateRuleIPv6 = False
                    continue

            if shouldCreateRuleIPv4:

                logging.info("Creating firewall rule '{}'"
                             .format(ruleIPv4Name))

                if self.plumbery.safeMode:
                    logging.info("- not in safe mode")

                else:

                    sourceIPv4 = DimensionDataFirewallAddress(
                        any_ip=False,
                        ip_address=source.private_ipv4_range_address,
                        ip_prefix_size=source.private_ipv4_range_size,
                        port_begin=None,
                        port_end=None)

                    ruleIPv4 = DimensionDataFirewallRule(
                        id=uuid4(),
                        action='ACCEPT_DECISIVELY',
                        name=ruleIPv4Name,
                        location=destination.location,
                        network_domain=destination.network_domain,
                        status='NORMAL',
                        ip_version='IPV4',
                        protocol='IP',
                        enabled='true',
                        source=sourceIPv4,
                        destination=destinationIPv4)

                    try:

                        self._ex_create_firewall_rule(
                            network_domain=destination.network_domain,
                            rule=ruleIPv4,
                            position='LAST')

                        logging.info("- in progress")

                    except Exception as feedback:

                        if 'NAME_NOT_UNIQUE' in str(feedback):
                            logging.info("- already there")

                        else:
                            logging.info("- unable to create firewall rule")
                            logging.error(str(feedback))

            if shouldCreateRuleIPv6:

                logging.info("Creating firewall rule '{}'"
                             .format(ruleIPv6Name))

                if self.plumbery.safeMode:
                    logging.info("- not in safe mode")

                else:

                    sourceIPv6 = DimensionDataFirewallAddress(
                        any_ip=False,
                        ip_address=source.ipv6_range_address,
                        ip_prefix_size=source.ipv6_range_size,
                        port_begin=None,
                        port_end=None)

                    ruleIPv6 = DimensionDataFirewallRule(
                        id=uuid4(),
                        action='ACCEPT_DECISIVELY',
                        name=ruleIPv6Name,
                        location=destination.location,
                        network_domain=destination.network_domain,
                        status='NORMAL',
                        ip_version='IPV6',
                        protocol='IP',
                        enabled='true',
                        source=sourceIPv6,
                        destination=destinationIPv6)

                    try:

                        self._ex_create_firewall_rule(
                            network_domain=destination.network_domain,
                            rule=ruleIPv6,
                            position='LAST')

                        logging.info("- in progress")

                    except Exception as feedback:

                        if 'NAME_NOT_UNIQUE' in str(feedback):
                            logging.info("- already there")

                        else:
                            logging.info("- unable to create firewall rule")
                            logging.error(str(feedback))

        ruleName = 'CCDEFAULT.DenyExternalInboundIPv6'
        for rule in self._list_firewall_rules():
            if rule.name.lower() == ruleName.lower():
                logging.info("Disabling firewall rule '{}'".format(ruleName))

                try:
                    if rule.enabled:
                        self.region.ex_set_firewall_rule_state(rule, False)
                        logging.info("- in progress")

                    else:
                        logging.info("- already there")

                except Exception as feedback:
                    logging.info("- unable to disable firewall rule")
                    logging.error(str(feedback))

        return True

    def _build_balancer(self):
        """
        Adds a load balancer for nodes in the blueprint

        Example in the fittings plan::

          - web:
              domain: *vdc1
              ethernet: *data
              nodes:
                - apache[10..19]
              listeners:
                - http:
                    port: 80
                    protocol: http
                - https:
                    port: 443
                    protocol: http
              balancer:
                algorithm: round_robin

        In this example, the load balancer is configured to accept web traffic
        and to distribute the workload across multiple web engines.

        One listener is configured for regular http protocol on port 80. The
        other listener is for secured web protocol, aka, https, on port 443.

        The algorithm used by default is ``round_robin``. This parameter
        can take any value among followings:

        * ``random``
        * ``round_robin``
        * ``least_connections``
        * ``weighted_round_robin``
        * ``weighted_least_connections``
        * ``shortest_response``
        * ``persistent_ip``

        """

        if 'listeners' not in self.blueprint:
            return True

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        driver = self.plumbery.get_balancer_driver(self.get_region_id())
        driver.ex_set_current_network_domain(domain.id)

        pool = self._get_pool()

        if pool is None:

            if 'balancer' in self.blueprint:
                settings = self.blueprint['balancer']
                if not isinstance(settings, dict):
                    settings = {}
            else:
                settings = {}

            name = self._name_pool()

            if 'algorithm' in settings:
                algorithm = settings['algorithm'].lower()
            else:
                algorithm = 'round_robin'

            algorithms = [
                'random',
                'round_robin',
                'least_connections',
                'weighted_round_robin',
                'weighted_least_connections',
                'shortest_response',
                'persistent_ip']

            if algorithm not in algorithms:
                raise PlumberyException(
                    "Error: unknown algorithm has been defined "
                    "for the pool '{}'!".format(name))

            logging.info("Creating pool '{}'".format(name))

            if self.plumbery.safeMode:
                logging.info("- not in safe mode")

            else:
                try:
                    pool = driver.ex_create_pool(
                        network_domain_id=domain.id,
                        name=name,
                        balancer_method=algorithm,
                        ex_description="#plumbery",
                        health_monitors=None,
                        service_down_action='NONE',
                        slow_ramp_time=30)

                    if self._cache_pools is None:
                        self._cache_pools = []
                    self._cache_pools.append(pool)

                    logging.info("- in progress")

                except Exception as feedback:

                    if 'NAME_NOT_UNIQUE' in str(feedback):
                        logging.info("- already there")

                    else:
                        logging.info("- unable to create pool")
                        logging.error(str(feedback))

        for item in self.blueprint['listeners']:

            if isinstance(item, dict):
                label = item.keys()[0]
                settings = item[label]
            else:
                label = str(item)
                settings = {}

            name = self.name_listener(label, settings)

            if self._get_listener(name):
                logging.info("Creating listener '{}'".format(name))
                logging.info("- already there")
                continue

            if 'port' in settings:
                port = settings['port']
            else:
                port = '80'

            port = int(port)
            if port < 1 or port > 65535:
                raise PlumberyException(
                    "Error: invalid port has been defined "
                    "for the listener '{}'!".format(label))

            if 'protocol' in settings:
                protocol = settings['protocol']
            else:
                protocol = 'http'

            protocols = ['http', 'https', 'tcp', 'udp']

            if protocol not in protocols:
                raise PlumberyException(
                    "Error: unknown protocol has been defined "
                    "for the listener '{}'!".format(label))

            logging.info("Creating listener '{}'".format(name))

            if self.plumbery.safeMode:
                logging.info("- not in safe mode")
                continue

            try:
                listener = driver.ex_create_virtual_listener(
                    network_domain_id=domain.id,
                    name=name,
                    ex_description="#plumbery",
                    port=port,
                    pool=pool,
                    persistence_profile=None,
                    fallback_persistence_profile=None,
                    irule=None,
                    protocol='TCP',
                    connection_limit=25000,
                    connection_rate_limit=2000,
                    source_port_preservation='PRESERVE')

                if self._cache_listeners is None:
                    self._cache_listeners = []
                self._cache_listeners.append(listener)

                logging.info("- in progress")

            except Exception as feedback:

                if 'NAME_NOT_UNIQUE' in str(feedback):
                    logging.info("- already there")

                elif 'NO_IP_ADDRESS_AVAILABLE' in str(feedback):
                    logging.info("- no more ipv4 address available "
                                 "-- assign more")

                else:
                    logging.info("- unable to create listener")
                    logging.error(str(feedback))

        return True

    def _destroy_balancer(self):
        """
        Destroys load balancer

        """

        if 'listeners' not in self.blueprint:
            return True

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        driver = self.plumbery.get_balancer_driver(self.get_region_id())
        driver.ex_set_current_network_domain(domain.id)

        for item in self.blueprint['listeners']:

            if isinstance(item, dict):
                label = item.keys()[0]
                settings = item[label]
            else:
                label = str(item)
                settings = {}

            name = self.name_listener(label, settings)

            listener = self._get_listener(name)

            logging.info("Destroying listener '{}'".format(name))

            if listener is None:
                logging.info("- not found")
                continue

            if self.plumbery.safeMode:
                logging.info("- not in safe mode")
                continue

            try:
                driver.destroy_balancer(listener)
                logging.info("- in progress")

            except Exception as feedback:

                if 'NOT_FOUND' in str(feedback):
                    logging.info("- not found")

                else:
                    logging.info("- unable to destroy listener")
                    logging.error(str(feedback))

        pool = self._get_pool()

        logging.info("Destroying pool '{}'".format(self._name_pool()))

        if pool is None:
            logging.info("- not found")

        elif self.plumbery.safeMode:
            logging.info("- not in safe mode")

        else:
            try:
                driver.ex_destroy_pool(pool)
                logging.info("- in progress")

            except Exception as feedback:

                if 'NAME_NOT_UNIQUE' in str(feedback):
                    logging.info("- already there")

                else:
                    logging.info("- unable to destroy pool")
                    logging.error(str(feedback))

    def name_listener(self, label, settings={}):
        return self.blueprint['target']                 \
            + '_' + self.facility.get_location_id().lower()      \
            + '.' + label + '.listener'

    def _get_listener(self, name):
        """
        Retrieves a listener attached to this blueprint

        """

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        driver = self.plumbery.get_balancer_driver(self.get_region_id())
        if driver is None:
            return None
        if domain is None:
            return None
        driver.ex_set_current_network_domain(domain.id)

        if self._cache_listeners is None:
            logging.info("Listing listeners")
            self._cache_listeners = driver.list_balancers()
            logging.info("- found {} listeners"
                         .format(len(self._cache_listeners)))

        for listener in self._cache_listeners:

            if listener.name.lower() == name.lower():
                return listener

        return None

    def _name_pool(self):
        return self.blueprint['target']                     \
            + '_' + self.facility.get_location_id().lower() + '.pool'

    def _get_pool(self):
        """
        Retrieves the pool attached to this blueprint

        """

        if 'listeners' not in self.blueprint:
            return None

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        driver = self.plumbery.get_balancer_driver(self.get_region_id())
        driver.ex_set_current_network_domain(domain.id)

        name = self._name_pool()

        if self._cache_pools is None:
            logging.info("Listing pools")
            self._cache_pools = driver.ex_get_pools()
            logging.info("- found {} pools".format(len(self._cache_pools)))

        for pool in self._cache_pools:

            if pool.name.lower() == name.lower():
                return pool

        return None

    def name_member(self, node):
        return node.name+'.plumbery'

    def _add_to_pool(self, node):
        """
        Makes a node a new member of the pool

        """

        if 'listeners' not in self.blueprint:
            return

        pool = self._get_pool()
        if pool is None:
            return

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        driver = self.plumbery.get_balancer_driver(self.get_region_id())
        driver.ex_set_current_network_domain(domain.id)

        logging.info("Adding '{}' to pool '{}'".format(node.name, pool.name))

        name = self.name_member(node)
        members = driver.ex_get_pool_members(pool.id)
        for member in members:

            if member.name == name:
                logging.info("- already there")
                return

        if self.plumbery.safeMode:
            logging.info("- not in safe mode")
            return

        try:
            member = driver.ex_create_node(
                network_domain_id=domain.id,
                name=name,
                ip=node.private_ips[0],
                ex_description='#plumbery')

            driver.ex_create_pool_member(
                pool=pool,
                node=member,
                port='*padding*')

            logging.info("- in progress")

        except Exception as feedback:

            if 'NAME_NOT_UNIQUE' in str(feedback):
                logging.info("- already there")
                logging.error(str(feedback))

            else:
                logging.info("- unable to add to pool")
                logging.error(str(feedback))

    def _remove_from_pool(self, node):
        """
        Removes a node from the pool

        """

        if 'listeners' not in self.blueprint:
            return

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        driver = self.plumbery.get_balancer_driver(self.get_region_id())
        driver.ex_set_current_network_domain(domain.id)

        pool = self._get_pool()
        if pool is not None:

            logging.info("Removing '{}' from pool '{}'".format(
                node.name,
                pool.name))

            members = driver.ex_get_pool_members(pool.id)

            found = False
            for member in members:

                if member.name == self.name_member(node):

                    if self.plumbery.safeMode:
                        logging.info("- not in safe mode")
                        return

                    try:
                        driver.balancer_detach_member(
                            balancer='*unused*',
                            member=member)
                        logging.info("- in progress")

                    except Exception as feedback:

                        if 'RESOURCE_NOT_FOUND' in str(feedback):
                            logging.info("- not found")

                        else:
                            logging.info("- unable to remove from pool")
                            logging.error(str(feedback))

                    found = True
                    break

            if not found:
                logging.info("- already there")

        try:
            logging.info("Destroying membership of '{}'".format(node.name))

            members = driver.ex_get_nodes()
            for member in members:
                if member.name == self.name_member(node):
                    driver.ex_destroy_node(member.id)
                    break

            logging.info("- in progress")

        except Exception as feedback:

            if 'RESOURCE_NOT_FOUND' in str(feedback):
                logging.info("- not found")

            else:
                logging.info("- unable to destroy membership")
                logging.error(str(feedback))

    def _destroy_firewall_rules(self):
        """
        Destroys firewall rules

        """

        if 'accept' not in self.blueprint['ethernet']:
            return True

        destinationLabel = self.blueprint['ethernet']['name']

        for item in self.blueprint['ethernet']['accept']:

            if isinstance(item, dict):
                label = item.keys()[0]
            else:
                label = str(item)

            sourceLabel = label.split('::').pop()

            ruleIPv4Name = self.name_firewall_rule(
                sourceLabel, destinationLabel, 'IP')

            ruleIPv6Name = self.name_firewall_rule(
                sourceLabel, destinationLabel, 'IPv6')

            for rule in self._list_firewall_rules():

                if rule.name == ruleIPv4Name or rule.name == ruleIPv6Name:

                    logging.info("Destroying firewall rule '{}'"
                                 .format(rule.name))

                    if self.plumbery.safeMode:
                        logging.info("- not in safe mode")

                    else:
                        try:
                            self.region.ex_delete_firewall_rule(rule)
                            logging.info("- in progress")

                        except Exception as feedback:

                            if 'RESOURCE_NOT_FOUND' in str(feedback):
                                logging.info("- not found")

                            else:
                                logging.info("- unable to destroy "
                                             "firewall rule")
                                logging.error(str(feedback))

    def destroy_blueprint(self, blueprint):
        """
        Destroys network and security elements of a blueprint

        :param blueprint: the various attributes of the target fittings
        :type blueprint: ``dict``

        This function looks after following service elements:

        * it releases public IPv4 addresses
        * it destroys firewall rules
        * it destroys the Ethernet network
        * it destroys the network domain

        The destruction is tentative, meaning that if the Ethernet network or
        the network domain have some dependency then they cannot be destroyed.
        This is happenign quite often since multiple blueprints can share the
        same Ethernet network or the same network domain.

        """

        self.blueprint = blueprint

        if ('domain' not in blueprint
                or type(blueprint['domain']) is not dict):

            raise PlumberyException(
                "Error: no network domain has been defined "
                "for the blueprint '{}'!".format(blueprint['target']))

        if ('ethernet' not in blueprint
                or type(blueprint['ethernet']) is not dict):

            raise PlumberyException(
                "Error: no ethernet network has been defined "
                "for the blueprint '{}'!".format(blueprint['target']))

        domainName = blueprint['domain']['name']
        networkName = blueprint['ethernet']['name']

        domain = self.get_network_domain(domainName)
        if domain is None:
            logging.info("Destroying Ethernet network '{}'"
                         .format(networkName))
            logging.info("- not found")
            logging.info("Destroying network domain '{}'".format(domainName))
            logging.info("- not found")
            return

        self._destroy_firewall_rules()

        self._destroy_balancer()

        self._release_ipv4()

        logging.info("Destroying Ethernet network '{}'".format(networkName))

        network = self.get_ethernet(networkName)
        if network is None:
            logging.info("- not found")

        elif ('destroy' in blueprint['ethernet']
                and blueprint['ethernet']['destroy'] == 'never'):
            logging.info("- this network can never be destroyed")

        elif self.plumbery.safeMode:
            logging.info("- not in safe mode")

        else:

            retry = True
            while True:
                try:
                    self.region.ex_delete_vlan(vlan=network)
                    logging.info("- in progress")

                    while True:
                        try:
                            time.sleep(10)
                            self.region.ex_get_vlan(vlan_id=network.id)
                        except Exception as feedback:
                            if 'RESOURCE_NOT_FOUND' in str(feedback):
                                break

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'RESOURCE_NOT_FOUND' in str(feedback):
                        logging.info("- not found")

                    elif 'HAS_DEPENDENCY' in str(feedback):

                        # give time to ensure nodes have been deleted
                        if retry:
                            retry = False
                            time.sleep(30)
                            continue

                        logging.info("- not now - stuff on it")
                        return

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        logging.info("- not now - locked")
                        logging.info(feedback)
                        return

                    else:
                        logging.info("- unable to destroy Ethernet network")
                        logging.error(str(feedback))
                        return

                break

        logging.info("Destroying network domain '{}'".format(domainName))

        if self.plumbery.safeMode:
            logging.info("- not in safe mode")
            return

        while True:
            try:
                self.region.ex_delete_network_domain(network_domain=domain)
                logging.info("- in progress")

            except Exception as feedback:

                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'RESOURCE_NOT_FOUND' in str(feedback):
                    logging.info("- not found")

                elif 'HAS_DEPENDENCY' in str(feedback):
                    logging.info("- not now - stuff on it")
                    return

                elif 'RESOURCE_LOCKED' in str(feedback):
                    logging.info("- not now - locked")
                    return

                else:
                    logging.info("- unable to destroy Ethernet network")
                    logging.error(str(feedback))
                    return

            break

    def _detach_node_from_internet(self, node):
        """
        Destroys address translation for one node

        :param node: node that was reachable from the internet
        :type node: :class:`libcloud.common.Node`

        """

        internal_ip = node.private_ips[0]
        domain = self.get_network_domain(self.blueprint['domain']['name'])
        for rule in self.region.ex_list_nat_rules(domain):
            if rule.internal_ip == internal_ip:

                logging.info("Detaching node '{}' from the internet"
                             .format(node.name))

                while True:
                    try:
                        self.region.ex_delete_nat_rule(rule)
                        logging.info("- in progress")

                    except Exception as feedback:
                        if 'RESOURCE_BUSY' in str(feedback):
                            time.sleep(10)
                            continue

                        elif 'RESOURCE_LOCKED' in str(feedback):
                            logging.info("- not now - locked")
                            return

                        else:
                            logging.info("- unable to remove "
                                         "address translation")
                            logging.error(str(feedback))

                    break

        for rule in self._list_firewall_rules():

            if rule.name.lower().startswith(node.name.lower()):

                logging.info("Destroying firewall rule '{}'"
                             .format(rule.name))

                if self.plumbery.safeMode:
                    logging.info("- not in safe mode")

                else:
                    self.region.ex_delete_firewall_rule(rule)
                    logging.info("- in progress")

    def _ex_create_firewall_rule(self, network_domain, rule, position):
        create_node = ET.Element('createFirewallRule', {'xmlns': TYPES_URN})
        ET.SubElement(create_node, "networkDomainId").text = network_domain.id
        ET.SubElement(create_node, "name").text = rule.name
        ET.SubElement(create_node, "action").text = rule.action
        ET.SubElement(create_node, "ipVersion").text = rule.ip_version
        ET.SubElement(create_node, "protocol").text = rule.protocol
        # Setup source port rule
        source = ET.SubElement(create_node, "source")
        source_ip = ET.SubElement(source, 'ip')
        if rule.source.any_ip:
            source_ip.set('address', 'ANY')
        else:
            source_ip.set('address', rule.source.ip_address)
            source_ip.set('prefixSize', str(rule.source.ip_prefix_size))
            if rule.source.port_begin is not None:
                source_port = ET.SubElement(source, 'port')
                source_port.set('begin', rule.source.port_begin)
            if rule.source.port_end is not None:
                source_port.set('end', rule.source.port_end)
        # Setup destination port rule
        dest = ET.SubElement(create_node, "destination")
        dest_ip = ET.SubElement(dest, 'ip')
        if rule.destination.any_ip:
            dest_ip.set('address', 'ANY')
        else:
            dest_ip.set('address', rule.destination.ip_address)
            if rule.destination.ip_prefix_size is not None:
                dest_ip.set('prefixSize', str(rule.destination.ip_prefix_size))
            if rule.destination.port_begin is not None:
                dest_port = ET.SubElement(dest, 'port')
                dest_port.set('begin', rule.destination.port_begin)
            if rule.destination.port_end is not None:
                dest_port.set('end', rule.destination.port_end)
        ET.SubElement(create_node, "enabled").text = 'true'
        placement = ET.SubElement(create_node, "placement")
        placement.set('position', position)

        response = self.region.connection.request_with_orgId_api_2(
            'network/createFirewallRule',
            method='POST',
            data=ET.tostring(create_node)).object

        rule_id = None
        for info in findall(response, 'info', TYPES_URN):
            if info.get('name') == 'firewallRuleId':
                rule_id = info.get('value')
        rule.id = rule_id
        return rule

    def get_container(self, blueprint):
        """
        Retrieves a domain attached to a blueprint

        :param blueprint: the various attributes of the target fittings
        :type blueprint: ``dict``

        :return: the infrastructure associated to the provided blueprint
        :rtype: :class:`plumbery.PlumberyInfrastructure` or `None``

        The returned object has at least a network domain and an Ethernet
        network, like in the following example::

            >>>container = infrastructure.get_container(blueprint)
            >>>print(container.domain.name)
            ...
            >>>print(container.network.name)
            ...

        """
        target = PlumberyInfrastructure(self.facility)

        target.blueprint = blueprint

        if ('domain' not in blueprint
                or type(blueprint['domain']) is not dict):

            raise PlumberyException(
                "Error: no network domain has been defined "
                "for the blueprint '{}'!".format(blueprint['target']))

        if ('ethernet' not in blueprint
                or type(blueprint['ethernet']) is not dict):

            raise PlumberyException(
                "Error: no ethernet network has been defined "
                "for the blueprint '{}'!".format(blueprint['target']))

        domainName = blueprint['domain']['name']
        target.domain = self.get_network_domain(domainName)

        networkName = blueprint['ethernet']['name']
        target.network = self.get_ethernet(networkName)

        return target

    def get_ethernet(self, path):
        """
        Retrieves an Ethernet network by name

        :param label: the name of the target Ethernet network
        :type label: ``str`` or ``list``of ``str``

        :return: an instance of an Ethernet network
        :rtype: :class:`VLAN` or ``None``

        This function searches firstly at the current facility. If the
        name is a complete path to a remote network, then plumbery looks
        there. If a different region is provided, then authentication is done
        against the related endpoint.

        For example if ``MyNetwork`` has been defined in a data centre in
        Europe::

            >>>infrastructure.get_ethernet('MyNetwork')
            >>>infrastructure.get_ethernet(['EU6', 'MyNetwork'])
            Looking for remote Ethernet network 'EU6::MyNetwork'
            - found it
            >>>infrastructure.get_ethernet(['dd-eu', 'EU6', 'MyNetwork'])
            Looking for offshore Ethernet network 'dd-eu::EU6::MyNetwork'
            - found it

        """

        if isinstance(path, str):
            path = [path]

        if len(path) == 1:

            if len(self.facility._cache_vlans) < 1:
                logging.info("Listing Ethernet networks")
                self.facility._cache_vlans = self.region.ex_list_vlans(
                    location=self.facility.location)
                logging.info("- found {} Ethernet networks"
                             .format(len(self.facility._cache_vlans)))

            for network in self.facility._cache_vlans:
                if network.name == path[0]:
                    return network

        elif len(path) == 2:

            if (len(self._cache_remote_vlan) == 3
                    and self._cache_remote_vlan[0] == path[0]
                    and self._cache_remote_vlan[1] == path[1]):

                return self._cache_remote_vlan[2]

            logging.info("Looking for remote Ethernet network '{}'"
                         .format('::'.join(path)))

            remoteLocation = self.region.ex_get_location_by_id(path[0])

            vlans = self.region.ex_list_vlans(location=remoteLocation)
            for network in vlans:
                if network.name == path[1]:
                    self._cache_remote_vlan += path
                    self._cache_remote_vlan.append(network)
                    logging.info("- found it")
                    return network

        elif len(path) == 3:

            if (len(self._cache_offshore_vlan) == 4
                    and self._cache_offshore_vlan[0] == path[0]
                    and self._cache_offshore_vlan[1] == path[1]
                    and self._cache_offshore_vlan[2] == path[2]):

                return self._cache_offshore_vlan[3]

            logging.info("Looking for offshore Ethernet network '{}'"
                         .format('::'.join(path)))

            offshore = self.plumbery.get_compute_driver(region=path[0])

            remoteLocation = offshore.ex_get_location_by_id(path[1])

            vlans = offshore.ex_list_vlans(location=remoteLocation)
            for network in vlans:
                if network.name == path[2]:
                    self._cache_offshore_vlan += path
                    self._cache_offshore_vlan.append(network)
                    logging.info("- found it")
                    return network

        return None

    def _get_ipv4(self):
        """
        Provides a free public IPv4 if possible

        """

        addresses = self._list_ipv4()
        if len(addresses) < 1:
            self._reserve_ipv4()
            addresses = self._list_ipv4()

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        if domain is None:
            return None

        for rule in self.region.ex_list_nat_rules(domain):
            addresses.remove(rule.external_ip)

        if 'balancer' in self.blueprint:
            addresses.pop(0)

        if len(addresses) < 1:
            return None

        return addresses[0]

    def get_network_domain(self, name):
        """
        Retrieves a network domain by name

        :param name: name of the target network domain
        :type name: ``str``

        """

        if len(self.facility._cache_network_domains) < 1:
            logging.info("Listing network domains")
            self.facility._cache_network_domains = \
                self.region.ex_list_network_domains(self.facility.location)
            logging.info("- found {} network domains"
                         .format(len(self.facility._cache_network_domains)))

        for domain in self.facility._cache_network_domains:
            if domain.name == name:
                return domain

        return None

    def _list_candidate_firewall_rules(self, node, ports=[]):
        """
        Lists rules that should apply to one node

        :param node: node that has to be reachable from the internet
        :type node: :class:`libcloud.common.Node`

        :param ports: the ports that have to be opened, or ``any``
        :type ports: a ``list`` of ``str``

        """

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        network = self.get_ethernet(self.blueprint['ethernet']['name'])

        internal_ip = node.private_ips[0]

        external_ip = None
        for rule in self.region.ex_list_nat_rules(domain):
            if rule.internal_ip == internal_ip:
                external_ip = rule.external_ip

        if external_ip is None:
            return {}

        candidates = {}

        if len(ports) < 1:
            ports = ['any']

        for port in ports:

            tokens = port.split('..')
            if len(tokens) == 1:
                if tokens[0].lower() == 'any':
                    port_begin = None
                else:
                    port_begin = tokens[0]
                port_end = None
                port = tokens[0]
            else:
                port_begin = tokens[0]
                port_end = tokens[1]
                port = port_begin+'..'+port_end

            ruleIPv4Name = self.name_firewall_rule(
                'Internet',
                node.name, 'TCPv4_'+port)

            sourceIPv4 = DimensionDataFirewallAddress(
                any_ip=True,
                ip_address=network.private_ipv4_range_address,
                ip_prefix_size=network.private_ipv4_range_size,
                port_begin=None,
                port_end=None)

            destinationIPv4 = DimensionDataFirewallAddress(
                any_ip=False,
                ip_address=external_ip,
                ip_prefix_size=None,
                port_begin=port_begin,
                port_end=port_end)

            ruleIPv4 = DimensionDataFirewallRule(
                id=uuid4(),
                action='ACCEPT_DECISIVELY',
                name=ruleIPv4Name,
                location=network.location,
                network_domain=network.network_domain,
                status='NORMAL',
                ip_version='IPV4',
                protocol='TCP',
                enabled='true',
                source=sourceIPv4,
                destination=destinationIPv4)

            candidates[ruleIPv4Name] = ruleIPv4

        return candidates

    def _list_firewall_rules(self):
        """
        Lists all existing rules for the current domain
        """
        if len(self._cache_firewall_rules) < 1:
            self._cache_firewall_rules = self.region.ex_list_firewall_rules(
                self.get_network_domain(self.blueprint['domain']['name']))

        return self._cache_firewall_rules

    def _list_ipv4(self):
        """
        Lists public IPv4 addresses that have been assigned to a domain

        :return: the full list of public IPv4 addresses assigned to the domain
        :rtype: ``list`` of ``str`` or ``[]``

        """

        addresses = []

        while True:
            try:
                blocks = self.region.ex_list_public_ip_blocks(
                    self.get_network_domain(self.blueprint['domain']['name']))
                for block in blocks:
                    splitted = block.base_ip.split('.')
                    for ticker in xrange(int(block.size)):
                        addresses.append('.'.join(splitted))
                        splitted[3] = str(int(splitted[3])+1)

            except Exception as feedback:

                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                else:
                    logging.info("Unable to list IPv4 public addresses")
                    logging.error(str(feedback))
                    return []

            break

        return addresses

    def name_firewall_rule(self, source, destination, protocol):
        """
        Provides a name for a firewall rule

        :param source: name of the source network
        :type source: ``str``

        :param destination: name of the destination network
        :type destination: ``str``

        :param protocol: the protocol that will flow
        :type protocol: ``str``

        Use this function to ensure consistent naming across firewall rules.

        Example::

            >>>source='gigafox.control'
            >>>destination='gigafox.production'
            >>>protocol='IP'
            >>>domain.name_firewall_rule(source, destination, protocol)
            'plumbery.FlowIPFromGigafoxControlToGigafoxProduction'

        """

        source = ''.join(e for e in source.title()
                         if e.isalnum() or e == '_')
        destination = ''.join(e for e in destination.title()
                              if e.isalnum() or e == '_')

        if source == 'Internet':
            return "{}.{}.plumbery".format(destination, protocol)

        else:
            return "From{}To{}.{}.plumbery".format(source,
                                                   destination,
                                                   protocol)

    def _release_ipv4(self):
        """
        Releases public IPv4 addresses assigned to the blueprint

        """

        domain = self.get_network_domain(self.blueprint['domain']['name'])
        if len(self.region.ex_list_nat_rules(domain)) > 0:
            return

        blocks = self.region.ex_list_public_ip_blocks(
            self.get_network_domain(self.blueprint['domain']['name']))

        if len(blocks) < 1:
            return

        logging.info('Releasing public IPv4 addresses')

        if self.plumbery.safeMode:
            logging.info("- not in safe mode")
            return

        for block in blocks:
            while True:
                try:
                    self.region.ex_delete_public_ip_block(block)
                    logging.info('- in progress')

                except Exception as feedback:

                    if 'RESOURCE_BUSY' in str(feedback):
                        time.sleep(10)
                        continue

                    elif 'HAS_DEPENDENCY' in str(feedback):
                        logging.info("- not now - stuff at '{}' and beyond"
                                     .format(block.base_ip))

                    elif 'RESOURCE_LOCKED' in str(feedback):
                        logging.info("- not now - locked")

                    else:
                        logging.info("- unable to release "
                                     "IPv4 public addresses ")
                        logging.error(str(feedback))

                break

    def _reserve_ipv4(self):
        """
        Reserves public addresses

        This function looks at current IPv4 addresses reserved for the
        target network domain, and adds more if needed.

        Example to reserve 8 IPv4 addresses in the fittings plan::

          - redis:
              domain:
                name: myVDC
                ipv4: 8

        """

        if 'ipv4' in self.blueprint['domain']:
            count = self.blueprint['domain']['ipv4']
        else:
            count = 2

        if count < 2 or count > 128:
            logging.info("Invalid count of requested IPv4 public addresses")
            return False

        actual = len(self._list_ipv4())
        if actual >= count:
            return True

        logging.info('Reserving additional public IPv4 addresses')

        if self.plumbery.safeMode:
            logging.info("- not in safe mode")
            return True

        while actual < count:
            try:
                block = self.region.ex_add_public_ip_block_to_network_domain(
                    self.get_network_domain(self.blueprint['domain']['name']))
                actual += int(block.size)
                logging.debug("- reserved {} addresses"
                              .format(int(block.size)))

            except Exception as feedback:

                if 'RESOURCE_BUSY' in str(feedback):
                    time.sleep(10)
                    continue

                elif 'RESOURCE_LOCKED' in str(feedback):
                    logging.info("- not now - locked")
                    return False

                # compensate for bug in Libcloud driver
                elif 'RESOURCE_NOT_FOUND' in str(feedback):
                    actual += 2
                    continue

                else:
                    logging.info("- unable to reserve IPv4 public addresses")
                    logging.error(str(feedback))
                    return False

        logging.info("- reserved {} addresses in total".format(actual))



#!/usr/bin/env python

"""
Tests for `plumbery` module.
"""

import logging
import socket
import unittest

from Crypto.PublicKey import RSA
import ast

from libcloud.common.types import InvalidCredsError

from plumbery.__main__ import parse_args, main
from plumbery.engine import PlumberyEngine
from plumbery import __version__

myPlan = """
---
safeMode: True
---
# Frankfurt in Europe
locationId: EU6
regionId: dd-eu

blueprints:

  - myBlueprint:
      domain:
        name: myDC
      ethernet:
        name: myVLAN
        subnet: 10.1.10.0
      nodes:
        - myServer
"""

myFacility = {
    'regionId': 'dd-eu',
    'locationId': 'EU7',
    'blueprints': [{
        'fake': {
            'domain': {
                'name': 'VDC1',
                'service': 'ADVANCED',
                'description': 'fake'},
            'ethernet': {
                'name': 'vlan1',
                'subnet': '10.0.10.0',
                'description': 'fake'},
            'nodes': [{
                'stackstorm': {
                    'description': 'fake',
                    'appliance': 'RedHat 6 64-bit 4 CPU'
                    }
                }]
            }
        }]
    }


class FakeLocation:

    id = 'EU7'
    name = 'data centre in Amsterdam'
    country = 'Netherlands'


class TestPlumberyEngine(unittest.TestCase):

    def test_set(self):

        settings = {
            'safeMode': False,
            'polishers': [
                {'ansible': {}},
                {'spit': {}},
                ]
            }

        self.engine = PlumberyEngine()
        self.engine.set_shared_secret('fake_secret')
        self.assertEqual(self.engine.get_shared_secret(), 'fake_secret')

        random = self.engine.get_secret('random')
        self.assertEqual(len(random), 9)
        self.assertEqual(self.engine.get_secret('random'), random)

        self.engine.set_user_name('fake_name')
        self.assertEqual(self.engine.get_user_name(), 'fake_name')

        self.engine.set_user_password('fake_password')
        self.assertEqual(self.engine.get_user_password(), 'fake_password')

        self.engine.set(settings)
        self.assertEqual(self.engine.safeMode, False)

        try:
            self.engine.from_text(myPlan)
            self.engine.add_facility(myFacility)
            self.assertEqual(len(self.engine.facilities), 2)

        except socket.gaierror:
            pass
        except InvalidCredsError:
            pass

    def test_lifecycle(self):

        self.engine = PlumberyEngine()
        self.engine.set_shared_secret('fake_secret')
        self.assertEqual(self.engine.get_shared_secret(), 'fake_secret')

        self.engine.set_user_name('fake_name')
        self.assertEqual(self.engine.get_user_name(), 'fake_name')

        self.engine.set_user_password('fake_password')
        self.assertEqual(self.engine.get_user_password(), 'fake_password')

        try:
            self.engine.do('build')
            self.engine.build_all_blueprints()
            self.engine.build_blueprint('myBlueprint')

            self.engine.do('start')
            self.engine.start_all_blueprints()
            self.engine.start_blueprint('myBlueprint')

            self.engine.do('polish')
            self.engine.polish_all_blueprints()
            self.engine.polish_blueprint('myBlueprint')

            self.engine.do('stop')
            self.engine.stop_all_blueprints()
            self.engine.stop_blueprint('myBlueprint')

            self.engine.wipe_all_blueprints()
            self.engine.wipe_blueprint('myBlueprint')

            self.engine.do('destroy')
            self.engine.destroy_all_blueprints()
            self.engine.destroy_blueprint('myBlueprint')

        except socket.gaierror:
            pass
        except InvalidCredsError:
            pass

    def test_lookup(self):

        self.engine = PlumberyEngine()
        self.assertEqual(self.engine.lookup('plumbery.version'), __version__)

        self.engine.secrets = {}
        random = self.engine.lookup('random.secret')
        self.assertEqual(len(random), 9)
        self.assertEqual(self.engine.lookup('random.secret'), random)

        md5 = self.engine.lookup('random.md5.secret')
        self.assertEqual(len(md5), 32)
        self.assertNotEqual(md5, random)

        sha = self.engine.lookup('random.sha1.secret')
        self.assertEqual(len(sha), 40)
        self.assertNotEqual(sha, random)

        sha = self.engine.lookup('random.sha256.secret')
        self.assertEqual(len(sha), 64)
        self.assertNotEqual(sha, random)

        self.engine.lookup('application.secret')
        self.engine.lookup('database.secret')
        self.engine.lookup('master.secret')
        self.engine.lookup('slave.secret')

        publicKey = self.engine.lookup('pair1.ssh.rsa_public')
        self.assertEqual(publicKey.startswith('ssh-rsa '), True)

        original = 'hello world'
        key = RSA.importKey(self.engine.lookup('pair1.rsa_public'))
        encrypted = key.publickey().encrypt(original, 32)

        privateKey = self.engine.lookup('pair1.rsa_private')
        self.assertEqual(privateKey.startswith(
            '-----BEGIN RSA PRIVATE KEY-----'), True)
        key = RSA.importKey(self.engine.lookup('pair1.rsa_private'))
        decrypted = key.decrypt(ast.literal_eval(str(encrypted)))
        self.assertEqual(decrypted, original)

        self.assertEqual(len(self.engine.secrets), 11)

    def test_parser(self):
        args = parse_args(['fittings.yaml', 'build', 'web'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'build')
        self.assertEqual(args.blueprints, ['web'])
        self.assertEqual(args.facilities, None)
        args = parse_args(['fittings.yaml', 'build', 'web', '-d'])
        self.assertEqual(
            logging.getLogger().getEffectiveLevel(), logging.DEBUG)
        args = parse_args(['fittings.yaml', 'build', 'web', '-q'])
        self.assertEqual(
            logging.getLogger().getEffectiveLevel(), logging.WARNING)
        args = parse_args(['fittings.yaml', 'start', '@NA12'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'start')
        self.assertEqual(args.blueprints, None)
        self.assertEqual(args.facilities, ['NA12'])
        args = parse_args([
            'fittings.yaml', 'rub', 'web', 'sql', '@NA9', '@NA12'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'rub')
        self.assertEqual(args.blueprints, ['web', 'sql'])
        self.assertEqual(args.facilities, ['NA9', 'NA12'])
        args = parse_args([
            'fittings.yaml', 'rub', 'web', '@NA9', 'sql', '@NA12'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'rub')
        self.assertEqual(args.blueprints, ['web', 'sql'])
        self.assertEqual(args.facilities, ['NA9', 'NA12'])
        args = parse_args(['fittings.yaml', 'polish'])
        self.assertEqual(args.fittings, 'fittings.yaml')
        self.assertEqual(args.action, 'polish')
        self.assertEqual(args.blueprints, None)
        self.assertEqual(args.facilities, None)

    def test_main(self):
        engine = PlumberyEngine()
        engine.from_text(myPlan)
        engine.set_user_name('fake_name')
        engine.set_user_password('fake_password')
        with self.assertRaises(SystemExit):
            main(['bad args'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'xyz123', 'web'], engine)
        with self.assertRaises(SystemExit):
            main(['-v'], engine)
        with self.assertRaises(SystemExit):
            main(['fittings.yaml', 'build', 'web', '-v'], engine)

if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())
